from typing import Iterator, List

import pandas as pd

from pypdtools.abc.scripts import AbstractEtl
from pypdtools.core.dataframe import PtDataFrame


class BaseEtl(AbstractEtl):
    """This class is used to create an ETL object."""

    def __init__(self) -> None:
        """This method is used to initialize the ETL object."""
        super().__init__()
        self.data = None
        self.executed = False

    def extract(self):
        """This method is used to get data from the source."""
        raise NotImplementedError

    def transform(self):
        """This method is used to transform the data."""
        raise NotImplementedError

    def load(self):
        """This method is used to load the data."""
        raise NotImplementedError

    def execute(self):
        """This method is used to execute the etl."""
        if self.executed:
            return self

        self.extract()
        self.transform()
        self.load()
        self.executed = True
        return self

    @property
    def data(self) -> pd.DataFrame:
        """This method is used to get the data."""
        self.execute()
        return self._df

    def __len__(self) -> int:
        """This method is used to get the length of the data."""
        self.execute()
        return super().__len__()

    def __iter__(self) -> Iterator:
        """This method is used to iterate over the data."""
        self.execute()
        return super().__iter__()

    def __aiter__(self) -> Iterator:
        """This method is used to async iterate over the data."""
        self.execute()
        return super().__aiter__()

    def __enter__(self) -> pd.DataFrame:
        """This method is used to enter the context."""
        self.execute()
        return super().__enter__()

    async def __aenter__(self) -> pd.DataFrame:
        """This method is used to async enter the context."""
        self.execute()
        return super().__aenter__()

    def __add__(self, other) -> PtDataFrame:
        """This method is used to add two dataframes."""
        self.execute()
        other.execute()
        return super().__add__(other)

    def __radd__(self, other) -> PtDataFrame:
        """This method is used to add two dataframes."""
        self.execute()
        other.execute()
        return super().__radd__(other)

    def col_to_list(self, col_name, drop_duplicates=True) -> List:
        """This method is used to get a column as a list."""
        self.execute()
        return super().col_to_list(col_name, drop_duplicates)



from abc import ABC, abstractmethod
from collections.abc import AsyncIterator, Awaitable, Iterator, Sized


class AbstractPtDataFrame(ABC, AsyncIterator, Awaitable, Sized, Iterator):
    """Abstract base class for DataFrame objects."""

    @abstractmethod
    def __repr__(self):
        ...

    @abstractmethod
    def __str__(self):
        ...

    @abstractmethod
    def __enter__(self):
        ...

    @abstractmethod
    async def __aenter__(self):
        ...

    @abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb):
        ...

    @abstractmethod
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        ...



from abc import abstractmethod

from pypdtools.core.dataframe import PtDataFrame


class AbstractEtl(PtDataFrame):
    """This class is used to create an ETL object.

    Args:
        PtDataFrame (_type_): _description_
    """

    @abstractmethod
    def extract(self):
        """This method is used to get data from the source."""
        ...

    @abstractmethod
    def transform(self):
        """This method is used to transform the data."""
        ...

    @abstractmethod
    def load(self):
        """This method is used to load the data."""
        ...

    @abstractmethod
    def execute(self):
        """This method is used to execute the etl."""
        ...



import asyncio
from typing import Awaitable, Iterator, List


class AsyncCollection:
    """A class that allows for asynchronous iteration over a collection.

    Args:
        collection (iterable): The collection to iterate over.

    Attributes:
        collection (iterable): The collection to iterate over.
    """

    def __init__(self, collection):
        self.collection = collection

    async def __aiter__(self) -> Iterator:
        """Asynchronously iterate over the collection.

        Yields:
            Any: The next item in the collection.
        """
        for item in self.collection:
            yield item
            await asyncio.sleep(0)

    async def __anext__(self) -> Awaitable:
        """Asynchronously get the next item in the collection.

        Raises:
            StopAsyncIteration: If there are no more items in the collection.

        Returns:
            Any: The next item in the collection.
        """
        try:
            item = next(self.iterator)
        except StopIteration:
            raise StopAsyncIteration
        else:
            await asyncio.sleep(0)
            return item

    async def __aenter__(self) -> "AsyncCollection":
        """Enter the async context.

        Returns:
            AsyncCollection: This object.
        """
        return self

    async def __aexit__(self, exc_type, exc, tb):
        """Exit the async context.

        Args:
            exc_type (type): The type of the exception (if any).
            exc (Exception): The exception instance (if any).
            tb (traceback): The traceback object (if any).
        """
        pass

    async def map(self, func) -> List:
        """Asynchronously apply a function to each item in the collection.

        Args:
            func (callable): The function to apply.

        Returns:
            List: A list of the results of applying the function to each item in the collection.
        """
        tasks = []
        for item in self.collection:
            tasks.append(asyncio.ensure_future(func(item)))
        results = await asyncio.gather(*tasks)
        return results



from asyncio import Lock, sleep
from functools import reduce
from typing import Any, Awaitable, Iterator, List

import pandas as pd

from pypdtools.abc.dataframe import AbstractPtDataFrame


class PtDataFrame(AbstractPtDataFrame):
    """
    A class representing a Pandas DataFrame with additional functionality.

    Args:
        df (pd.DataFrame, optional): The Pandas DataFrame to be encapsulated. Defaults to an empty DataFrame.
    """

    def __init__(self, df: pd.DataFrame = pd.DataFrame()) -> None:
        """
        Initialize the PtDataFrame.

        Args:
            df (pd.DataFrame, optional): The Pandas DataFrame to be encapsulated. Defaults to an empty DataFrame.
        """
        self._df = df
        self._iterator = self._df.itertuples(index=False)
        self._lock = Lock()

    @property
    def df(self) -> pd.DataFrame:
        """
        Get the encapsulated Pandas DataFrame.

        Returns:
            pd.DataFrame: The encapsulated Pandas DataFrame.
        """
        return self._df

    def __repr__(self) -> str:
        """
        Return a string representation of the PtDataFrame.

        Returns:
            str: A string representation of the PtDataFrame.
        """
        return self._df.__repr__()

    def __str__(self) -> str:
        """
        Return a string representation of the PtDataFrame.

        Returns:
            str: A string representation of the PtDataFrame.
        """
        return self._df.__str__()

    def __await__(self) -> Awaitable:
        """
        Awaitable support for the PtDataFrame.

        Returns:
            Awaitable: An awaitable object.
        """
        return sleep(0.0001).__await__()

    def __enter__(self) -> pd.DataFrame:
        """
        Enter the context manager and return the encapsulated Pandas DataFrame.

        Returns:
            pd.DataFrame: The encapsulated Pandas DataFrame.
        """
        return self._df

    async def __aenter__(self) -> pd.DataFrame:
        """
        Asynchronous version of __enter__(). Enter the context manager and return the encapsulated Pandas DataFrame.

        Returns:
            pd.DataFrame: The encapsulated Pandas DataFrame.
        """
        return self._df

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """
        Exit the context manager.

        Args:
            exc_type (type): The exception type, if any.
            exc_val (Exception): The exception instance, if any.
            exc_tb (traceback): The traceback, if any.
        """
        return None

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        """
        Asynchronous version of __exit__(). Exit the context manager.

        Args:
            exc_type (type): The exception type, if any.
            exc_val (Exception): The exception instance, if any.
            exc_tb (traceback): The traceback, if any.
        """
        return None

    def __aiter__(self) -> Iterator:
        """
        Asynchronous iterator support for the PtDataFrame.

        Returns:
            Iterator: An asynchronous iterator.
        """
        self._iterator = self._df.itertuples(index=False)

        async def _aiter() -> Iterator:
            """
            Asynchronous iterator helper function.

            Yields:
                Iterator: An item from the PtDataFrame.
            """
            while True:
                try:
                    yield self._iterator.__next__()
                except StopIteration:
                    break

        return _aiter()

    def __iter__(self) -> Iterator:
        """
               Iterator

        support for the PtDataFrame.

               Returns:
                   Iterator: An iterator.
        """
        self._iterator = self._df.itertuples(index=False)

        return (it for it in self._iterator)

    def __len__(self) -> int:
        """
        Return the length of the PtDataFrame.

        Returns:
            int: The length of the PtDataFrame.
        """
        return len(self._df)

    async def __anext__(self) -> Awaitable:
        """
        Asynchronous iterator support for the PtDataFrame.

        Returns:
            Awaitable: An awaitable object.
        """
        async with self._lock:
            try:
                return self._iterator.__next__()
            except StopIteration:
                raise StopAsyncIteration

    def __next__(self) -> Any:
        """
        Iterator support for the PtDataFrame.

        Returns:
            Any: An item from the PtDataFrame.
        """
        return self._iterator.__next__()

    def __add__(self, other) -> "PtDataFrame":
        """
        Concatenate the PtDataFrame with another PtDataFrame.

        Args:
            other (PtDataFrame): The PtDataFrame to concatenate.

        Returns:
            PtDataFrame: The concatenated PtDataFrame.
        """
        concat = pd.concat([self._df, other._df], ignore_index=True)
        return PtDataFrame(concat)

    def __iadd__(self, other) -> "PtDataFrame":
        """
        In-place concatenation of the PtDataFrame with another PtDataFrame.

        Args:
            other (PtDataFrame): The PtDataFrame to concatenate.

        Returns:
            PtDataFrame: The concatenated PtDataFrame.
        """
        self._df = pd.concat([self._df, other._df], ignore_index=True)
        return self

    @classmethod
    def reduce(cls, df_list) -> "PtDataFrame":
        """
        Reduce a list of PtDataFrame objects into a single PtDataFrame.

        Args:
            df_list (List[PtDataFrame]): The list of PtDataFrame objects to reduce.

        Returns:
            PtDataFrame: The reduced PtDataFrame.
        """
        return PtDataFrame(reduce(lambda x, y: x + y, df_list).df)

    def col_to_list(self, col_name, drop_duplicates=True) -> List:
        """
        Extract a column as a list from the PtDataFrame.

        Args:
            col_name (str): The name of the column to extract.
            drop_duplicates (bool, optional): Whether to drop duplicate values. Defaults to True.

        Returns:
            List: The column values as a list.
        """
        if drop_duplicates:
            return self._df[col_name].drop_duplicates().tolist()
        return self._df[col_name].tolist()
